from pyDOE import *
from doepy.pydoe_corrected import fullfact_corrected,fracfact_corrected, fracfact_by_res,bbdesign_corrected, ccdesign_corrected
from diversipy import *
import pandas as pd
import numpy as np

name = "doepy"